/* --------------------------------------------------------------------------

    MIT License

    Copyright (c) 2020-2024 Jurek Muszynski (rekmus)

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

-----------------------------------------------------------------------------

    Table access base class
    nodepp.org

-------------------------------------------------------------------------- */

#ifndef NPP_MYSQL_H
#define NPP_MYSQL_H

#ifndef EOS
#define EOS                             (char)0         /* End Of String */
#endif

#define CDB_ISOLATION_DIRTY_READ        1               /* isolation levels */
#define CDB_ISOLATION_COMMITTED_READ    2
#define CDB_ISOLATION_REPEATABLE_READ   3

#define CDB_LOCK_MODE_NOT_WAIT          1               /* lock modes */
#define CDB_LOCK_MODE_WAIT              2
#define CDB_LOCK_MODE_WAIT_1            3
#define CDB_LOCK_MODE_WAIT_2            4
#define CDB_LOCK_MODE_WAIT_3            5
#define CDB_LOCK_MODE_WAIT_5            6
#define CDB_LOCK_MODE_WAIT_10           7
#define CDB_LOCK_MODE_WAIT_15           8

/* customizable */

#ifndef CDB_MAX_INSTANCES
#define CDB_MAX_INSTANCES               100             /* max available instances per (final) type */
#endif

#ifndef CDB_DEFAULT_FIRST
#define CDB_DEFAULT_FIRST               1000            /* read only first n records by default */
#endif

#ifndef CDB_STR_LEN
#define CDB_STR_LEN                     255             /* search string length */
#endif

#define CDB_STR_BUF_LEN                 256             /* search string buffer size */

#ifndef CDB_SQLBUF
#define CDB_SQLBUF                      16384           /* SQL statement buffer length */
#endif

#ifndef CDB_DEFAULT_PORT
#define CDB_DEFAULT_PORT                3306            /* default MySQL port */
#endif


#include <stdio.h>
#include <string.h>
#include <iostream>
#include <mysql.h>


#if MYSQL_VERSION_ID == 80031
typedef bool my_bool;
#endif


#ifndef SSL_MODE_DISABLED
#define SSL_MODE_DISABLED               1
#endif

#ifndef CR_SSL_CONNECTION_ERROR
#define CR_SSL_CONNECTION_ERROR         2026
#endif


class Cdb
{
public:
    Cdb();
    ~Cdb();

    // The statics below are pretty much self-explanatory
    static void DBOpen(const std::string& dbName, const std::string& user="", const std::string& password="", const std::string& host="", int port=0);
    static void DBClose();

    static void SetIsolation(int level);
    static void SetLockMode(int mode);

    static void Begin();
    static void Commit();
    static void Rollback();

    // EXECUTE any SQL query without preparing
    static void Execute(const std::string& query);

    // Compose message & throw exception
    static void ThrowSQL(const char *action);

    // Set SKIP SQL query option
    void SetSkip(unsigned skip);

    // Set FIRST SQL query option
    void SetFirst(unsigned first);

    // Define WHERE clause & PREPARE SQL statement
    void PrepareCursor(const std::string& where="");

    // Define WHERE clause & PREPARE SQL statement
    void PrepareCount(const std::string& where="");

    // Reset all values
    virtual void Reset() = 0;

    // ----------------------------------------------------------------------------------
    // DECLARE & OPEN SQL cursor based on PREPAREd statement
    // ----------------------------------------------------------------------------------
    void OpenCursor();

    void OpenCursor(int);
    void OpenCursor(unsigned);
    void OpenCursor(float);
    void OpenCursor(double);
    void OpenCursor(const std::string&);

    void OpenCursor(int, int);
    void OpenCursor(int, unsigned);
    void OpenCursor(int, float);
    void OpenCursor(int, double);
    void OpenCursor(int, const std::string&);
    void OpenCursor(unsigned, int);
    void OpenCursor(unsigned, unsigned);
    void OpenCursor(unsigned, float);
    void OpenCursor(unsigned, double);
    void OpenCursor(unsigned, const std::string&);
    void OpenCursor(float, int);
    void OpenCursor(float, unsigned);
    void OpenCursor(float, float);
    void OpenCursor(float, double);
    void OpenCursor(float, const std::string&);
    void OpenCursor(double, int);
    void OpenCursor(double, unsigned);
    void OpenCursor(double, float);
    void OpenCursor(double, double);
    void OpenCursor(double, const std::string&);
    void OpenCursor(const std::string&, int);
    void OpenCursor(const std::string&, unsigned);
    void OpenCursor(const std::string&, float);
    void OpenCursor(const std::string&, double);
    void OpenCursor(const std::string&, const std::string&);

#ifdef CDB_3_ARGS

    void OpenCursor(int, int, int);
    void OpenCursor(int, int, unsigned);
    void OpenCursor(int, int, float);
    void OpenCursor(int, int, double);
    void OpenCursor(int, int, const std::string&);
    void OpenCursor(int, unsigned, int);
    void OpenCursor(int, unsigned, unsigned);
    void OpenCursor(int, unsigned, float);
    void OpenCursor(int, unsigned, double);
    void OpenCursor(int, unsigned, const std::string&);
    void OpenCursor(int, float, int);
    void OpenCursor(int, float, unsigned);
    void OpenCursor(int, float, float);
    void OpenCursor(int, float, double);
    void OpenCursor(int, float, const std::string&);
    void OpenCursor(int, double, int);
    void OpenCursor(int, double, unsigned);
    void OpenCursor(int, double, float);
    void OpenCursor(int, double, double);
    void OpenCursor(int, double, const std::string&);
    void OpenCursor(int, const std::string&, int);
    void OpenCursor(int, const std::string&, unsigned);
    void OpenCursor(int, const std::string&, float);
    void OpenCursor(int, const std::string&, double);
    void OpenCursor(int, const std::string&, const std::string&);
    void OpenCursor(unsigned, int, int);
    void OpenCursor(unsigned, int, unsigned);
    void OpenCursor(unsigned, int, float);
    void OpenCursor(unsigned, int, double);
    void OpenCursor(unsigned, int, const std::string&);
    void OpenCursor(unsigned, unsigned, int);
    void OpenCursor(unsigned, unsigned, unsigned);
    void OpenCursor(unsigned, unsigned, float);
    void OpenCursor(unsigned, unsigned, double);
    void OpenCursor(unsigned, unsigned, const std::string&);
    void OpenCursor(unsigned, float, int);
    void OpenCursor(unsigned, float, unsigned);
    void OpenCursor(unsigned, float, float);
    void OpenCursor(unsigned, float, double);
    void OpenCursor(unsigned, float, const std::string&);
    void OpenCursor(unsigned, double, int);
    void OpenCursor(unsigned, double, unsigned);
    void OpenCursor(unsigned, double, float);
    void OpenCursor(unsigned, double, double);
    void OpenCursor(unsigned, double, const std::string&);
    void OpenCursor(unsigned, const std::string&, int);
    void OpenCursor(unsigned, const std::string&, unsigned);
    void OpenCursor(unsigned, const std::string&, float);
    void OpenCursor(unsigned, const std::string&, double);
    void OpenCursor(unsigned, const std::string&, const std::string&);
    void OpenCursor(float, int, int);
    void OpenCursor(float, int, unsigned);
    void OpenCursor(float, int, float);
    void OpenCursor(float, int, double);
    void OpenCursor(float, int, const std::string&);
    void OpenCursor(float, unsigned, int);
    void OpenCursor(float, unsigned, unsigned);
    void OpenCursor(float, unsigned, float);
    void OpenCursor(float, unsigned, double);
    void OpenCursor(float, unsigned, const std::string&);
    void OpenCursor(float, float, int);
    void OpenCursor(float, float, unsigned);
    void OpenCursor(float, float, float);
    void OpenCursor(float, float, double);
    void OpenCursor(float, float, const std::string&);
    void OpenCursor(float, double, int);
    void OpenCursor(float, double, unsigned);
    void OpenCursor(float, double, float);
    void OpenCursor(float, double, double);
    void OpenCursor(float, double, const std::string&);
    void OpenCursor(float, const std::string&, int);
    void OpenCursor(float, const std::string&, unsigned);
    void OpenCursor(float, const std::string&, float);
    void OpenCursor(float, const std::string&, double);
    void OpenCursor(float, const std::string&, const std::string&);
    void OpenCursor(double, int, int);
    void OpenCursor(double, int, unsigned);
    void OpenCursor(double, int, float);
    void OpenCursor(double, int, double);
    void OpenCursor(double, int, const std::string&);
    void OpenCursor(double, unsigned, int);
    void OpenCursor(double, unsigned, unsigned);
    void OpenCursor(double, unsigned, float);
    void OpenCursor(double, unsigned, double);
    void OpenCursor(double, unsigned, const std::string&);
    void OpenCursor(double, float, int);
    void OpenCursor(double, float, unsigned);
    void OpenCursor(double, float, float);
    void OpenCursor(double, float, double);
    void OpenCursor(double, float, const std::string&);
    void OpenCursor(double, double, int);
    void OpenCursor(double, double, unsigned);
    void OpenCursor(double, double, float);
    void OpenCursor(double, double, double);
    void OpenCursor(double, double, const std::string&);
    void OpenCursor(double, const std::string&, int);
    void OpenCursor(double, const std::string&, unsigned);
    void OpenCursor(double, const std::string&, float);
    void OpenCursor(double, const std::string&, double);
    void OpenCursor(double, const std::string&, const std::string&);
    void OpenCursor(const std::string&, int, int);
    void OpenCursor(const std::string&, int, unsigned);
    void OpenCursor(const std::string&, int, float);
    void OpenCursor(const std::string&, int, double);
    void OpenCursor(const std::string&, int, const std::string&);
    void OpenCursor(const std::string&, unsigned, int);
    void OpenCursor(const std::string&, unsigned, unsigned);
    void OpenCursor(const std::string&, unsigned, float);
    void OpenCursor(const std::string&, unsigned, double);
    void OpenCursor(const std::string&, unsigned, const std::string&);
    void OpenCursor(const std::string&, float, int);
    void OpenCursor(const std::string&, float, unsigned);
    void OpenCursor(const std::string&, float, float);
    void OpenCursor(const std::string&, float, double);
    void OpenCursor(const std::string&, float, const std::string&);
    void OpenCursor(const std::string&, double, int);
    void OpenCursor(const std::string&, double, unsigned);
    void OpenCursor(const std::string&, double, float);
    void OpenCursor(const std::string&, double, double);
    void OpenCursor(const std::string&, double, const std::string&);
    void OpenCursor(const std::string&, const std::string&, int);
    void OpenCursor(const std::string&, const std::string&, unsigned);
    void OpenCursor(const std::string&, const std::string&, float);
    void OpenCursor(const std::string&, const std::string&, double);
    void OpenCursor(const std::string&, const std::string&, const std::string&);

#endif   /* CDB_3_ARGS */

#ifdef CDB_4_ARGS

    void OpenCursor(int, int, int, int);
    void OpenCursor(int, int, int, unsigned);
    void OpenCursor(int, int, int, float);
    void OpenCursor(int, int, int, double);
    void OpenCursor(int, int, int, const std::string&);
    void OpenCursor(int, int, unsigned, int);
    void OpenCursor(int, int, unsigned, unsigned);
    void OpenCursor(int, int, unsigned, float);
    void OpenCursor(int, int, unsigned, double);
    void OpenCursor(int, int, unsigned, const std::string&);
    void OpenCursor(int, int, float, int);
    void OpenCursor(int, int, float, unsigned);
    void OpenCursor(int, int, float, float);
    void OpenCursor(int, int, float, double);
    void OpenCursor(int, int, float, const std::string&);
    void OpenCursor(int, int, double, int);
    void OpenCursor(int, int, double, unsigned);
    void OpenCursor(int, int, double, float);
    void OpenCursor(int, int, double, double);
    void OpenCursor(int, int, double, const std::string&);
    void OpenCursor(int, int, const std::string&, int);
    void OpenCursor(int, int, const std::string&, unsigned);
    void OpenCursor(int, int, const std::string&, float);
    void OpenCursor(int, int, const std::string&, double);
    void OpenCursor(int, int, const std::string&, const std::string&);
    void OpenCursor(int, unsigned, int, int);
    void OpenCursor(int, unsigned, int, unsigned);
    void OpenCursor(int, unsigned, int, float);
    void OpenCursor(int, unsigned, int, double);
    void OpenCursor(int, unsigned, int, const std::string&);
    void OpenCursor(int, unsigned, unsigned, int);
    void OpenCursor(int, unsigned, unsigned, unsigned);
    void OpenCursor(int, unsigned, unsigned, float);
    void OpenCursor(int, unsigned, unsigned, double);
    void OpenCursor(int, unsigned, unsigned, const std::string&);
    void OpenCursor(int, unsigned, float, int);
    void OpenCursor(int, unsigned, float, unsigned);
    void OpenCursor(int, unsigned, float, float);
    void OpenCursor(int, unsigned, float, double);
    void OpenCursor(int, unsigned, float, const std::string&);
    void OpenCursor(int, unsigned, double, int);
    void OpenCursor(int, unsigned, double, unsigned);
    void OpenCursor(int, unsigned, double, float);
    void OpenCursor(int, unsigned, double, double);
    void OpenCursor(int, unsigned, double, const std::string&);
    void OpenCursor(int, unsigned, const std::string&, int);
    void OpenCursor(int, unsigned, const std::string&, unsigned);
    void OpenCursor(int, unsigned, const std::string&, float);
    void OpenCursor(int, unsigned, const std::string&, double);
    void OpenCursor(int, unsigned, const std::string&, const std::string&);
    void OpenCursor(int, float, int, int);
    void OpenCursor(int, float, int, unsigned);
    void OpenCursor(int, float, int, float);
    void OpenCursor(int, float, int, double);
    void OpenCursor(int, float, int, const std::string&);
    void OpenCursor(int, float, unsigned, int);
    void OpenCursor(int, float, unsigned, unsigned);
    void OpenCursor(int, float, unsigned, float);
    void OpenCursor(int, float, unsigned, double);
    void OpenCursor(int, float, unsigned, const std::string&);
    void OpenCursor(int, float, float, int);
    void OpenCursor(int, float, float, unsigned);
    void OpenCursor(int, float, float, float);
    void OpenCursor(int, float, float, double);
    void OpenCursor(int, float, float, const std::string&);
    void OpenCursor(int, float, double, int);
    void OpenCursor(int, float, double, unsigned);
    void OpenCursor(int, float, double, float);
    void OpenCursor(int, float, double, double);
    void OpenCursor(int, float, double, const std::string&);
    void OpenCursor(int, float, const std::string&, int);
    void OpenCursor(int, float, const std::string&, unsigned);
    void OpenCursor(int, float, const std::string&, float);
    void OpenCursor(int, float, const std::string&, double);
    void OpenCursor(int, float, const std::string&, const std::string&);
    void OpenCursor(int, double, int, int);
    void OpenCursor(int, double, int, unsigned);
    void OpenCursor(int, double, int, float);
    void OpenCursor(int, double, int, double);
    void OpenCursor(int, double, int, const std::string&);
    void OpenCursor(int, double, unsigned, int);
    void OpenCursor(int, double, unsigned, unsigned);
    void OpenCursor(int, double, unsigned, float);
    void OpenCursor(int, double, unsigned, double);
    void OpenCursor(int, double, unsigned, const std::string&);
    void OpenCursor(int, double, float, int);
    void OpenCursor(int, double, float, unsigned);
    void OpenCursor(int, double, float, float);
    void OpenCursor(int, double, float, double);
    void OpenCursor(int, double, float, const std::string&);
    void OpenCursor(int, double, double, int);
    void OpenCursor(int, double, double, unsigned);
    void OpenCursor(int, double, double, float);
    void OpenCursor(int, double, double, double);
    void OpenCursor(int, double, double, const std::string&);
    void OpenCursor(int, double, const std::string&, int);
    void OpenCursor(int, double, const std::string&, unsigned);
    void OpenCursor(int, double, const std::string&, float);
    void OpenCursor(int, double, const std::string&, double);
    void OpenCursor(int, double, const std::string&, const std::string&);
    void OpenCursor(int, const std::string&, int, int);
    void OpenCursor(int, const std::string&, int, unsigned);
    void OpenCursor(int, const std::string&, int, float);
    void OpenCursor(int, const std::string&, int, double);
    void OpenCursor(int, const std::string&, int, const std::string&);
    void OpenCursor(int, const std::string&, unsigned, int);
    void OpenCursor(int, const std::string&, unsigned, unsigned);
    void OpenCursor(int, const std::string&, unsigned, float);
    void OpenCursor(int, const std::string&, unsigned, double);
    void OpenCursor(int, const std::string&, unsigned, const std::string&);
    void OpenCursor(int, const std::string&, float, int);
    void OpenCursor(int, const std::string&, float, unsigned);
    void OpenCursor(int, const std::string&, float, float);
    void OpenCursor(int, const std::string&, float, double);
    void OpenCursor(int, const std::string&, float, const std::string&);
    void OpenCursor(int, const std::string&, double, int);
    void OpenCursor(int, const std::string&, double, unsigned);
    void OpenCursor(int, const std::string&, double, float);
    void OpenCursor(int, const std::string&, double, double);
    void OpenCursor(int, const std::string&, double, const std::string&);
    void OpenCursor(int, const std::string&, const std::string&, int);
    void OpenCursor(int, const std::string&, const std::string&, unsigned);
    void OpenCursor(int, const std::string&, const std::string&, float);
    void OpenCursor(int, const std::string&, const std::string&, double);
    void OpenCursor(int, const std::string&, const std::string&, const std::string&);
    void OpenCursor(unsigned, int, int, int);
    void OpenCursor(unsigned, int, int, unsigned);
    void OpenCursor(unsigned, int, int, float);
    void OpenCursor(unsigned, int, int, double);
    void OpenCursor(unsigned, int, int, const std::string&);
    void OpenCursor(unsigned, int, unsigned, int);
    void OpenCursor(unsigned, int, unsigned, unsigned);
    void OpenCursor(unsigned, int, unsigned, float);
    void OpenCursor(unsigned, int, unsigned, double);
    void OpenCursor(unsigned, int, unsigned, const std::string&);
    void OpenCursor(unsigned, int, float, int);
    void OpenCursor(unsigned, int, float, unsigned);
    void OpenCursor(unsigned, int, float, float);
    void OpenCursor(unsigned, int, float, double);
    void OpenCursor(unsigned, int, float, const std::string&);
    void OpenCursor(unsigned, int, double, int);
    void OpenCursor(unsigned, int, double, unsigned);
    void OpenCursor(unsigned, int, double, float);
    void OpenCursor(unsigned, int, double, double);
    void OpenCursor(unsigned, int, double, const std::string&);
    void OpenCursor(unsigned, int, const std::string&, int);
    void OpenCursor(unsigned, int, const std::string&, unsigned);
    void OpenCursor(unsigned, int, const std::string&, float);
    void OpenCursor(unsigned, int, const std::string&, double);
    void OpenCursor(unsigned, int, const std::string&, const std::string&);
    void OpenCursor(unsigned, unsigned, int, int);
    void OpenCursor(unsigned, unsigned, int, unsigned);
    void OpenCursor(unsigned, unsigned, int, float);
    void OpenCursor(unsigned, unsigned, int, double);
    void OpenCursor(unsigned, unsigned, int, const std::string&);
    void OpenCursor(unsigned, unsigned, unsigned, int);
    void OpenCursor(unsigned, unsigned, unsigned, unsigned);
    void OpenCursor(unsigned, unsigned, unsigned, float);
    void OpenCursor(unsigned, unsigned, unsigned, double);
    void OpenCursor(unsigned, unsigned, unsigned, const std::string&);
    void OpenCursor(unsigned, unsigned, float, int);
    void OpenCursor(unsigned, unsigned, float, unsigned);
    void OpenCursor(unsigned, unsigned, float, float);
    void OpenCursor(unsigned, unsigned, float, double);
    void OpenCursor(unsigned, unsigned, float, const std::string&);
    void OpenCursor(unsigned, unsigned, double, int);
    void OpenCursor(unsigned, unsigned, double, unsigned);
    void OpenCursor(unsigned, unsigned, double, float);
    void OpenCursor(unsigned, unsigned, double, double);
    void OpenCursor(unsigned, unsigned, double, const std::string&);
    void OpenCursor(unsigned, unsigned, const std::string&, int);
    void OpenCursor(unsigned, unsigned, const std::string&, unsigned);
    void OpenCursor(unsigned, unsigned, const std::string&, float);
    void OpenCursor(unsigned, unsigned, const std::string&, double);
    void OpenCursor(unsigned, unsigned, const std::string&, const std::string&);
    void OpenCursor(unsigned, float, int, int);
    void OpenCursor(unsigned, float, int, unsigned);
    void OpenCursor(unsigned, float, int, float);
    void OpenCursor(unsigned, float, int, double);
    void OpenCursor(unsigned, float, int, const std::string&);
    void OpenCursor(unsigned, float, unsigned, int);
    void OpenCursor(unsigned, float, unsigned, unsigned);
    void OpenCursor(unsigned, float, unsigned, float);
    void OpenCursor(unsigned, float, unsigned, double);
    void OpenCursor(unsigned, float, unsigned, const std::string&);
    void OpenCursor(unsigned, float, float, int);
    void OpenCursor(unsigned, float, float, unsigned);
    void OpenCursor(unsigned, float, float, float);
    void OpenCursor(unsigned, float, float, double);
    void OpenCursor(unsigned, float, float, const std::string&);
    void OpenCursor(unsigned, float, double, int);
    void OpenCursor(unsigned, float, double, unsigned);
    void OpenCursor(unsigned, float, double, float);
    void OpenCursor(unsigned, float, double, double);
    void OpenCursor(unsigned, float, double, const std::string&);
    void OpenCursor(unsigned, float, const std::string&, int);
    void OpenCursor(unsigned, float, const std::string&, unsigned);
    void OpenCursor(unsigned, float, const std::string&, float);
    void OpenCursor(unsigned, float, const std::string&, double);
    void OpenCursor(unsigned, float, const std::string&, const std::string&);
    void OpenCursor(unsigned, double, int, int);
    void OpenCursor(unsigned, double, int, unsigned);
    void OpenCursor(unsigned, double, int, float);
    void OpenCursor(unsigned, double, int, double);
    void OpenCursor(unsigned, double, int, const std::string&);
    void OpenCursor(unsigned, double, unsigned, int);
    void OpenCursor(unsigned, double, unsigned, unsigned);
    void OpenCursor(unsigned, double, unsigned, float);
    void OpenCursor(unsigned, double, unsigned, double);
    void OpenCursor(unsigned, double, unsigned, const std::string&);
    void OpenCursor(unsigned, double, float, int);
    void OpenCursor(unsigned, double, float, unsigned);
    void OpenCursor(unsigned, double, float, float);
    void OpenCursor(unsigned, double, float, double);
    void OpenCursor(unsigned, double, float, const std::string&);
    void OpenCursor(unsigned, double, double, int);
    void OpenCursor(unsigned, double, double, unsigned);
    void OpenCursor(unsigned, double, double, float);
    void OpenCursor(unsigned, double, double, double);
    void OpenCursor(unsigned, double, double, const std::string&);
    void OpenCursor(unsigned, double, const std::string&, int);
    void OpenCursor(unsigned, double, const std::string&, unsigned);
    void OpenCursor(unsigned, double, const std::string&, float);
    void OpenCursor(unsigned, double, const std::string&, double);
    void OpenCursor(unsigned, double, const std::string&, const std::string&);
    void OpenCursor(unsigned, const std::string&, int, int);
    void OpenCursor(unsigned, const std::string&, int, unsigned);
    void OpenCursor(unsigned, const std::string&, int, float);
    void OpenCursor(unsigned, const std::string&, int, double);
    void OpenCursor(unsigned, const std::string&, int, const std::string&);
    void OpenCursor(unsigned, const std::string&, unsigned, int);
    void OpenCursor(unsigned, const std::string&, unsigned, unsigned);
    void OpenCursor(unsigned, const std::string&, unsigned, float);
    void OpenCursor(unsigned, const std::string&, unsigned, double);
    void OpenCursor(unsigned, const std::string&, unsigned, const std::string&);
    void OpenCursor(unsigned, const std::string&, float, int);
    void OpenCursor(unsigned, const std::string&, float, unsigned);
    void OpenCursor(unsigned, const std::string&, float, float);
    void OpenCursor(unsigned, const std::string&, float, double);
    void OpenCursor(unsigned, const std::string&, float, const std::string&);
    void OpenCursor(unsigned, const std::string&, double, int);
    void OpenCursor(unsigned, const std::string&, double, unsigned);
    void OpenCursor(unsigned, const std::string&, double, float);
    void OpenCursor(unsigned, const std::string&, double, double);
    void OpenCursor(unsigned, const std::string&, double, const std::string&);
    void OpenCursor(unsigned, const std::string&, const std::string&, int);
    void OpenCursor(unsigned, const std::string&, const std::string&, unsigned);
    void OpenCursor(unsigned, const std::string&, const std::string&, float);
    void OpenCursor(unsigned, const std::string&, const std::string&, double);
    void OpenCursor(unsigned, const std::string&, const std::string&, const std::string&);
    void OpenCursor(float, int, int, int);
    void OpenCursor(float, int, int, unsigned);
    void OpenCursor(float, int, int, float);
    void OpenCursor(float, int, int, double);
    void OpenCursor(float, int, int, const std::string&);
    void OpenCursor(float, int, unsigned, int);
    void OpenCursor(float, int, unsigned, unsigned);
    void OpenCursor(float, int, unsigned, float);
    void OpenCursor(float, int, unsigned, double);
    void OpenCursor(float, int, unsigned, const std::string&);
    void OpenCursor(float, int, float, int);
    void OpenCursor(float, int, float, unsigned);
    void OpenCursor(float, int, float, float);
    void OpenCursor(float, int, float, double);
    void OpenCursor(float, int, float, const std::string&);
    void OpenCursor(float, int, double, int);
    void OpenCursor(float, int, double, unsigned);
    void OpenCursor(float, int, double, float);
    void OpenCursor(float, int, double, double);
    void OpenCursor(float, int, double, const std::string&);
    void OpenCursor(float, int, const std::string&, int);
    void OpenCursor(float, int, const std::string&, unsigned);
    void OpenCursor(float, int, const std::string&, float);
    void OpenCursor(float, int, const std::string&, double);
    void OpenCursor(float, int, const std::string&, const std::string&);
    void OpenCursor(float, unsigned, int, int);
    void OpenCursor(float, unsigned, int, unsigned);
    void OpenCursor(float, unsigned, int, float);
    void OpenCursor(float, unsigned, int, double);
    void OpenCursor(float, unsigned, int, const std::string&);
    void OpenCursor(float, unsigned, unsigned, int);
    void OpenCursor(float, unsigned, unsigned, unsigned);
    void OpenCursor(float, unsigned, unsigned, float);
    void OpenCursor(float, unsigned, unsigned, double);
    void OpenCursor(float, unsigned, unsigned, const std::string&);
    void OpenCursor(float, unsigned, float, int);
    void OpenCursor(float, unsigned, float, unsigned);
    void OpenCursor(float, unsigned, float, float);
    void OpenCursor(float, unsigned, float, double);
    void OpenCursor(float, unsigned, float, const std::string&);
    void OpenCursor(float, unsigned, double, int);
    void OpenCursor(float, unsigned, double, unsigned);
    void OpenCursor(float, unsigned, double, float);
    void OpenCursor(float, unsigned, double, double);
    void OpenCursor(float, unsigned, double, const std::string&);
    void OpenCursor(float, unsigned, const std::string&, int);
    void OpenCursor(float, unsigned, const std::string&, unsigned);
    void OpenCursor(float, unsigned, const std::string&, float);
    void OpenCursor(float, unsigned, const std::string&, double);
    void OpenCursor(float, unsigned, const std::string&, const std::string&);
    void OpenCursor(float, float, int, int);
    void OpenCursor(float, float, int, unsigned);
    void OpenCursor(float, float, int, float);
    void OpenCursor(float, float, int, double);
    void OpenCursor(float, float, int, const std::string&);
    void OpenCursor(float, float, unsigned, int);
    void OpenCursor(float, float, unsigned, unsigned);
    void OpenCursor(float, float, unsigned, float);
    void OpenCursor(float, float, unsigned, double);
    void OpenCursor(float, float, unsigned, const std::string&);
    void OpenCursor(float, float, float, int);
    void OpenCursor(float, float, float, unsigned);
    void OpenCursor(float, float, float, float);
    void OpenCursor(float, float, float, double);
    void OpenCursor(float, float, float, const std::string&);
    void OpenCursor(float, float, double, int);
    void OpenCursor(float, float, double, unsigned);
    void OpenCursor(float, float, double, float);
    void OpenCursor(float, float, double, double);
    void OpenCursor(float, float, double, const std::string&);
    void OpenCursor(float, float, const std::string&, int);
    void OpenCursor(float, float, const std::string&, unsigned);
    void OpenCursor(float, float, const std::string&, float);
    void OpenCursor(float, float, const std::string&, double);
    void OpenCursor(float, float, const std::string&, const std::string&);
    void OpenCursor(float, double, int, int);
    void OpenCursor(float, double, int, unsigned);
    void OpenCursor(float, double, int, float);
    void OpenCursor(float, double, int, double);
    void OpenCursor(float, double, int, const std::string&);
    void OpenCursor(float, double, unsigned, int);
    void OpenCursor(float, double, unsigned, unsigned);
    void OpenCursor(float, double, unsigned, float);
    void OpenCursor(float, double, unsigned, double);
    void OpenCursor(float, double, unsigned, const std::string&);
    void OpenCursor(float, double, float, int);
    void OpenCursor(float, double, float, unsigned);
    void OpenCursor(float, double, float, float);
    void OpenCursor(float, double, float, double);
    void OpenCursor(float, double, float, const std::string&);
    void OpenCursor(float, double, double, int);
    void OpenCursor(float, double, double, unsigned);
    void OpenCursor(float, double, double, float);
    void OpenCursor(float, double, double, double);
    void OpenCursor(float, double, double, const std::string&);
    void OpenCursor(float, double, const std::string&, int);
    void OpenCursor(float, double, const std::string&, unsigned);
    void OpenCursor(float, double, const std::string&, float);
    void OpenCursor(float, double, const std::string&, double);
    void OpenCursor(float, double, const std::string&, const std::string&);
    void OpenCursor(float, const std::string&, int, int);
    void OpenCursor(float, const std::string&, int, unsigned);
    void OpenCursor(float, const std::string&, int, float);
    void OpenCursor(float, const std::string&, int, double);
    void OpenCursor(float, const std::string&, int, const std::string&);
    void OpenCursor(float, const std::string&, unsigned, int);
    void OpenCursor(float, const std::string&, unsigned, unsigned);
    void OpenCursor(float, const std::string&, unsigned, float);
    void OpenCursor(float, const std::string&, unsigned, double);
    void OpenCursor(float, const std::string&, unsigned, const std::string&);
    void OpenCursor(float, const std::string&, float, int);
    void OpenCursor(float, const std::string&, float, unsigned);
    void OpenCursor(float, const std::string&, float, float);
    void OpenCursor(float, const std::string&, float, double);
    void OpenCursor(float, const std::string&, float, const std::string&);
    void OpenCursor(float, const std::string&, double, int);
    void OpenCursor(float, const std::string&, double, unsigned);
    void OpenCursor(float, const std::string&, double, float);
    void OpenCursor(float, const std::string&, double, double);
    void OpenCursor(float, const std::string&, double, const std::string&);
    void OpenCursor(float, const std::string&, const std::string&, int);
    void OpenCursor(float, const std::string&, const std::string&, unsigned);
    void OpenCursor(float, const std::string&, const std::string&, float);
    void OpenCursor(float, const std::string&, const std::string&, double);
    void OpenCursor(float, const std::string&, const std::string&, const std::string&);
    void OpenCursor(double, int, int, int);
    void OpenCursor(double, int, int, unsigned);
    void OpenCursor(double, int, int, float);
    void OpenCursor(double, int, int, double);
    void OpenCursor(double, int, int, const std::string&);
    void OpenCursor(double, int, unsigned, int);
    void OpenCursor(double, int, unsigned, unsigned);
    void OpenCursor(double, int, unsigned, float);
    void OpenCursor(double, int, unsigned, double);
    void OpenCursor(double, int, unsigned, const std::string&);
    void OpenCursor(double, int, float, int);
    void OpenCursor(double, int, float, unsigned);
    void OpenCursor(double, int, float, float);
    void OpenCursor(double, int, float, double);
    void OpenCursor(double, int, float, const std::string&);
    void OpenCursor(double, int, double, int);
    void OpenCursor(double, int, double, unsigned);
    void OpenCursor(double, int, double, float);
    void OpenCursor(double, int, double, double);
    void OpenCursor(double, int, double, const std::string&);
    void OpenCursor(double, int, const std::string&, int);
    void OpenCursor(double, int, const std::string&, unsigned);
    void OpenCursor(double, int, const std::string&, float);
    void OpenCursor(double, int, const std::string&, double);
    void OpenCursor(double, int, const std::string&, const std::string&);
    void OpenCursor(double, unsigned, int, int);
    void OpenCursor(double, unsigned, int, unsigned);
    void OpenCursor(double, unsigned, int, float);
    void OpenCursor(double, unsigned, int, double);
    void OpenCursor(double, unsigned, int, const std::string&);
    void OpenCursor(double, unsigned, unsigned, int);
    void OpenCursor(double, unsigned, unsigned, unsigned);
    void OpenCursor(double, unsigned, unsigned, float);
    void OpenCursor(double, unsigned, unsigned, double);
    void OpenCursor(double, unsigned, unsigned, const std::string&);
    void OpenCursor(double, unsigned, float, int);
    void OpenCursor(double, unsigned, float, unsigned);
    void OpenCursor(double, unsigned, float, float);
    void OpenCursor(double, unsigned, float, double);
    void OpenCursor(double, unsigned, float, const std::string&);
    void OpenCursor(double, unsigned, double, int);
    void OpenCursor(double, unsigned, double, unsigned);
    void OpenCursor(double, unsigned, double, float);
    void OpenCursor(double, unsigned, double, double);
    void OpenCursor(double, unsigned, double, const std::string&);
    void OpenCursor(double, unsigned, const std::string&, int);
    void OpenCursor(double, unsigned, const std::string&, unsigned);
    void OpenCursor(double, unsigned, const std::string&, float);
    void OpenCursor(double, unsigned, const std::string&, double);
    void OpenCursor(double, unsigned, const std::string&, const std::string&);
    void OpenCursor(double, float, int, int);
    void OpenCursor(double, float, int, unsigned);
    void OpenCursor(double, float, int, float);
    void OpenCursor(double, float, int, double);
    void OpenCursor(double, float, int, const std::string&);
    void OpenCursor(double, float, unsigned, int);
    void OpenCursor(double, float, unsigned, unsigned);
    void OpenCursor(double, float, unsigned, float);
    void OpenCursor(double, float, unsigned, double);
    void OpenCursor(double, float, unsigned, const std::string&);
    void OpenCursor(double, float, float, int);
    void OpenCursor(double, float, float, unsigned);
    void OpenCursor(double, float, float, float);
    void OpenCursor(double, float, float, double);
    void OpenCursor(double, float, float, const std::string&);
    void OpenCursor(double, float, double, int);
    void OpenCursor(double, float, double, unsigned);
    void OpenCursor(double, float, double, float);
    void OpenCursor(double, float, double, double);
    void OpenCursor(double, float, double, const std::string&);
    void OpenCursor(double, float, const std::string&, int);
    void OpenCursor(double, float, const std::string&, unsigned);
    void OpenCursor(double, float, const std::string&, float);
    void OpenCursor(double, float, const std::string&, double);
    void OpenCursor(double, float, const std::string&, const std::string&);
    void OpenCursor(double, double, int, int);
    void OpenCursor(double, double, int, unsigned);
    void OpenCursor(double, double, int, float);
    void OpenCursor(double, double, int, double);
    void OpenCursor(double, double, int, const std::string&);
    void OpenCursor(double, double, unsigned, int);
    void OpenCursor(double, double, unsigned, unsigned);
    void OpenCursor(double, double, unsigned, float);
    void OpenCursor(double, double, unsigned, double);
    void OpenCursor(double, double, unsigned, const std::string&);
    void OpenCursor(double, double, float, int);
    void OpenCursor(double, double, float, unsigned);
    void OpenCursor(double, double, float, float);
    void OpenCursor(double, double, float, double);
    void OpenCursor(double, double, float, const std::string&);
    void OpenCursor(double, double, double, int);
    void OpenCursor(double, double, double, unsigned);
    void OpenCursor(double, double, double, float);
    void OpenCursor(double, double, double, double);
    void OpenCursor(double, double, double, const std::string&);
    void OpenCursor(double, double, const std::string&, int);
    void OpenCursor(double, double, const std::string&, unsigned);
    void OpenCursor(double, double, const std::string&, float);
    void OpenCursor(double, double, const std::string&, double);
    void OpenCursor(double, double, const std::string&, const std::string&);
    void OpenCursor(double, const std::string&, int, int);
    void OpenCursor(double, const std::string&, int, unsigned);
    void OpenCursor(double, const std::string&, int, float);
    void OpenCursor(double, const std::string&, int, double);
    void OpenCursor(double, const std::string&, int, const std::string&);
    void OpenCursor(double, const std::string&, unsigned, int);
    void OpenCursor(double, const std::string&, unsigned, unsigned);
    void OpenCursor(double, const std::string&, unsigned, float);
    void OpenCursor(double, const std::string&, unsigned, double);
    void OpenCursor(double, const std::string&, unsigned, const std::string&);
    void OpenCursor(double, const std::string&, float, int);
    void OpenCursor(double, const std::string&, float, unsigned);
    void OpenCursor(double, const std::string&, float, float);
    void OpenCursor(double, const std::string&, float, double);
    void OpenCursor(double, const std::string&, float, const std::string&);
    void OpenCursor(double, const std::string&, double, int);
    void OpenCursor(double, const std::string&, double, unsigned);
    void OpenCursor(double, const std::string&, double, float);
    void OpenCursor(double, const std::string&, double, double);
    void OpenCursor(double, const std::string&, double, const std::string&);
    void OpenCursor(double, const std::string&, const std::string&, int);
    void OpenCursor(double, const std::string&, const std::string&, unsigned);
    void OpenCursor(double, const std::string&, const std::string&, float);
    void OpenCursor(double, const std::string&, const std::string&, double);
    void OpenCursor(double, const std::string&, const std::string&, const std::string&);
    void OpenCursor(const std::string&, int, int, int);
    void OpenCursor(const std::string&, int, int, unsigned);
    void OpenCursor(const std::string&, int, int, float);
    void OpenCursor(const std::string&, int, int, double);
    void OpenCursor(const std::string&, int, int, const std::string&);
    void OpenCursor(const std::string&, int, unsigned, int);
    void OpenCursor(const std::string&, int, unsigned, unsigned);
    void OpenCursor(const std::string&, int, unsigned, float);
    void OpenCursor(const std::string&, int, unsigned, double);
    void OpenCursor(const std::string&, int, unsigned, const std::string&);
    void OpenCursor(const std::string&, int, float, int);
    void OpenCursor(const std::string&, int, float, unsigned);
    void OpenCursor(const std::string&, int, float, float);
    void OpenCursor(const std::string&, int, float, double);
    void OpenCursor(const std::string&, int, float, const std::string&);
    void OpenCursor(const std::string&, int, double, int);
    void OpenCursor(const std::string&, int, double, unsigned);
    void OpenCursor(const std::string&, int, double, float);
    void OpenCursor(const std::string&, int, double, double);
    void OpenCursor(const std::string&, int, double, const std::string&);
    void OpenCursor(const std::string&, int, const std::string&, int);
    void OpenCursor(const std::string&, int, const std::string&, unsigned);
    void OpenCursor(const std::string&, int, const std::string&, float);
    void OpenCursor(const std::string&, int, const std::string&, double);
    void OpenCursor(const std::string&, int, const std::string&, const std::string&);
    void OpenCursor(const std::string&, unsigned, int, int);
    void OpenCursor(const std::string&, unsigned, int, unsigned);
    void OpenCursor(const std::string&, unsigned, int, float);
    void OpenCursor(const std::string&, unsigned, int, double);
    void OpenCursor(const std::string&, unsigned, int, const std::string&);
    void OpenCursor(const std::string&, unsigned, unsigned, int);
    void OpenCursor(const std::string&, unsigned, unsigned, unsigned);
    void OpenCursor(const std::string&, unsigned, unsigned, float);
    void OpenCursor(const std::string&, unsigned, unsigned, double);
    void OpenCursor(const std::string&, unsigned, unsigned, const std::string&);
    void OpenCursor(const std::string&, unsigned, float, int);
    void OpenCursor(const std::string&, unsigned, float, unsigned);
    void OpenCursor(const std::string&, unsigned, float, float);
    void OpenCursor(const std::string&, unsigned, float, double);
    void OpenCursor(const std::string&, unsigned, float, const std::string&);
    void OpenCursor(const std::string&, unsigned, double, int);
    void OpenCursor(const std::string&, unsigned, double, unsigned);
    void OpenCursor(const std::string&, unsigned, double, float);
    void OpenCursor(const std::string&, unsigned, double, double);
    void OpenCursor(const std::string&, unsigned, double, const std::string&);
    void OpenCursor(const std::string&, unsigned, const std::string&, int);
    void OpenCursor(const std::string&, unsigned, const std::string&, unsigned);
    void OpenCursor(const std::string&, unsigned, const std::string&, float);
    void OpenCursor(const std::string&, unsigned, const std::string&, double);
    void OpenCursor(const std::string&, unsigned, const std::string&, const std::string&);
    void OpenCursor(const std::string&, float, int, int);
    void OpenCursor(const std::string&, float, int, unsigned);
    void OpenCursor(const std::string&, float, int, float);
    void OpenCursor(const std::string&, float, int, double);
    void OpenCursor(const std::string&, float, int, const std::string&);
    void OpenCursor(const std::string&, float, unsigned, int);
    void OpenCursor(const std::string&, float, unsigned, unsigned);
    void OpenCursor(const std::string&, float, unsigned, float);
    void OpenCursor(const std::string&, float, unsigned, double);
    void OpenCursor(const std::string&, float, unsigned, const std::string&);
    void OpenCursor(const std::string&, float, float, int);
    void OpenCursor(const std::string&, float, float, unsigned);
    void OpenCursor(const std::string&, float, float, float);
    void OpenCursor(const std::string&, float, float, double);
    void OpenCursor(const std::string&, float, float, const std::string&);
    void OpenCursor(const std::string&, float, double, int);
    void OpenCursor(const std::string&, float, double, unsigned);
    void OpenCursor(const std::string&, float, double, float);
    void OpenCursor(const std::string&, float, double, double);
    void OpenCursor(const std::string&, float, double, const std::string&);
    void OpenCursor(const std::string&, float, const std::string&, int);
    void OpenCursor(const std::string&, float, const std::string&, unsigned);
    void OpenCursor(const std::string&, float, const std::string&, float);
    void OpenCursor(const std::string&, float, const std::string&, double);
    void OpenCursor(const std::string&, float, const std::string&, const std::string&);
    void OpenCursor(const std::string&, double, int, int);
    void OpenCursor(const std::string&, double, int, unsigned);
    void OpenCursor(const std::string&, double, int, float);
    void OpenCursor(const std::string&, double, int, double);
    void OpenCursor(const std::string&, double, int, const std::string&);
    void OpenCursor(const std::string&, double, unsigned, int);
    void OpenCursor(const std::string&, double, unsigned, unsigned);
    void OpenCursor(const std::string&, double, unsigned, float);
    void OpenCursor(const std::string&, double, unsigned, double);
    void OpenCursor(const std::string&, double, unsigned, const std::string&);
    void OpenCursor(const std::string&, double, float, int);
    void OpenCursor(const std::string&, double, float, unsigned);
    void OpenCursor(const std::string&, double, float, float);
    void OpenCursor(const std::string&, double, float, double);
    void OpenCursor(const std::string&, double, float, const std::string&);
    void OpenCursor(const std::string&, double, double, int);
    void OpenCursor(const std::string&, double, double, unsigned);
    void OpenCursor(const std::string&, double, double, float);
    void OpenCursor(const std::string&, double, double, double);
    void OpenCursor(const std::string&, double, double, const std::string&);
    void OpenCursor(const std::string&, double, const std::string&, int);
    void OpenCursor(const std::string&, double, const std::string&, unsigned);
    void OpenCursor(const std::string&, double, const std::string&, float);
    void OpenCursor(const std::string&, double, const std::string&, double);
    void OpenCursor(const std::string&, double, const std::string&, const std::string&);
    void OpenCursor(const std::string&, const std::string&, int, int);
    void OpenCursor(const std::string&, const std::string&, int, unsigned);
    void OpenCursor(const std::string&, const std::string&, int, float);
    void OpenCursor(const std::string&, const std::string&, int, double);
    void OpenCursor(const std::string&, const std::string&, int, const std::string&);
    void OpenCursor(const std::string&, const std::string&, unsigned, int);
    void OpenCursor(const std::string&, const std::string&, unsigned, unsigned);
    void OpenCursor(const std::string&, const std::string&, unsigned, float);
    void OpenCursor(const std::string&, const std::string&, unsigned, double);
    void OpenCursor(const std::string&, const std::string&, unsigned, const std::string&);
    void OpenCursor(const std::string&, const std::string&, float, int);
    void OpenCursor(const std::string&, const std::string&, float, unsigned);
    void OpenCursor(const std::string&, const std::string&, float, float);
    void OpenCursor(const std::string&, const std::string&, float, double);
    void OpenCursor(const std::string&, const std::string&, float, const std::string&);
    void OpenCursor(const std::string&, const std::string&, double, int);
    void OpenCursor(const std::string&, const std::string&, double, unsigned);
    void OpenCursor(const std::string&, const std::string&, double, float);
    void OpenCursor(const std::string&, const std::string&, double, double);
    void OpenCursor(const std::string&, const std::string&, double, const std::string&);
    void OpenCursor(const std::string&, const std::string&, const std::string&, int);
    void OpenCursor(const std::string&, const std::string&, const std::string&, unsigned);
    void OpenCursor(const std::string&, const std::string&, const std::string&, float);
    void OpenCursor(const std::string&, const std::string&, const std::string&, double);
    void OpenCursor(const std::string&, const std::string&, const std::string&, const std::string&);

#endif   /* CDB_4_ARGS */



    // ----------------------------------------------------------------------------------
    // EXECUTE PREPAREd COUNT(*) statement
    // ----------------------------------------------------------------------------------
    unsigned GetCount();

    unsigned GetCount(int);
    unsigned GetCount(unsigned);
    unsigned GetCount(float);
    unsigned GetCount(double);
    unsigned GetCount(const std::string&);

    unsigned GetCount(int, int);
    unsigned GetCount(int, unsigned);
    unsigned GetCount(int, float);
    unsigned GetCount(int, double);
    unsigned GetCount(int, const std::string&);
    unsigned GetCount(unsigned, int);
    unsigned GetCount(unsigned, unsigned);
    unsigned GetCount(unsigned, float);
    unsigned GetCount(unsigned, double);
    unsigned GetCount(unsigned, const std::string&);
    unsigned GetCount(float, int);
    unsigned GetCount(float, unsigned);
    unsigned GetCount(float, float);
    unsigned GetCount(float, double);
    unsigned GetCount(float, const std::string&);
    unsigned GetCount(double, int);
    unsigned GetCount(double, unsigned);
    unsigned GetCount(double, float);
    unsigned GetCount(double, double);
    unsigned GetCount(double, const std::string&);
    unsigned GetCount(const std::string&, int);
    unsigned GetCount(const std::string&, unsigned);
    unsigned GetCount(const std::string&, float);
    unsigned GetCount(const std::string&, double);
    unsigned GetCount(const std::string&, const std::string&);

#ifdef CDB_3_ARGS

    unsigned GetCount(int, int, int);
    unsigned GetCount(int, int, unsigned);
    unsigned GetCount(int, int, float);
    unsigned GetCount(int, int, double);
    unsigned GetCount(int, int, const std::string&);
    unsigned GetCount(int, unsigned, int);
    unsigned GetCount(int, unsigned, unsigned);
    unsigned GetCount(int, unsigned, float);
    unsigned GetCount(int, unsigned, double);
    unsigned GetCount(int, unsigned, const std::string&);
    unsigned GetCount(int, float, int);
    unsigned GetCount(int, float, unsigned);
    unsigned GetCount(int, float, float);
    unsigned GetCount(int, float, double);
    unsigned GetCount(int, float, const std::string&);
    unsigned GetCount(int, double, int);
    unsigned GetCount(int, double, unsigned);
    unsigned GetCount(int, double, float);
    unsigned GetCount(int, double, double);
    unsigned GetCount(int, double, const std::string&);
    unsigned GetCount(int, const std::string&, int);
    unsigned GetCount(int, const std::string&, unsigned);
    unsigned GetCount(int, const std::string&, float);
    unsigned GetCount(int, const std::string&, double);
    unsigned GetCount(int, const std::string&, const std::string&);
    unsigned GetCount(unsigned, int, int);
    unsigned GetCount(unsigned, int, unsigned);
    unsigned GetCount(unsigned, int, float);
    unsigned GetCount(unsigned, int, double);
    unsigned GetCount(unsigned, int, const std::string&);
    unsigned GetCount(unsigned, unsigned, int);
    unsigned GetCount(unsigned, unsigned, unsigned);
    unsigned GetCount(unsigned, unsigned, float);
    unsigned GetCount(unsigned, unsigned, double);
    unsigned GetCount(unsigned, unsigned, const std::string&);
    unsigned GetCount(unsigned, float, int);
    unsigned GetCount(unsigned, float, unsigned);
    unsigned GetCount(unsigned, float, float);
    unsigned GetCount(unsigned, float, double);
    unsigned GetCount(unsigned, float, const std::string&);
    unsigned GetCount(unsigned, double, int);
    unsigned GetCount(unsigned, double, unsigned);
    unsigned GetCount(unsigned, double, float);
    unsigned GetCount(unsigned, double, double);
    unsigned GetCount(unsigned, double, const std::string&);
    unsigned GetCount(unsigned, const std::string&, int);
    unsigned GetCount(unsigned, const std::string&, unsigned);
    unsigned GetCount(unsigned, const std::string&, float);
    unsigned GetCount(unsigned, const std::string&, double);
    unsigned GetCount(unsigned, const std::string&, const std::string&);
    unsigned GetCount(float, int, int);
    unsigned GetCount(float, int, unsigned);
    unsigned GetCount(float, int, float);
    unsigned GetCount(float, int, double);
    unsigned GetCount(float, int, const std::string&);
    unsigned GetCount(float, unsigned, int);
    unsigned GetCount(float, unsigned, unsigned);
    unsigned GetCount(float, unsigned, float);
    unsigned GetCount(float, unsigned, double);
    unsigned GetCount(float, unsigned, const std::string&);
    unsigned GetCount(float, float, int);
    unsigned GetCount(float, float, unsigned);
    unsigned GetCount(float, float, float);
    unsigned GetCount(float, float, double);
    unsigned GetCount(float, float, const std::string&);
    unsigned GetCount(float, double, int);
    unsigned GetCount(float, double, unsigned);
    unsigned GetCount(float, double, float);
    unsigned GetCount(float, double, double);
    unsigned GetCount(float, double, const std::string&);
    unsigned GetCount(float, const std::string&, int);
    unsigned GetCount(float, const std::string&, unsigned);
    unsigned GetCount(float, const std::string&, float);
    unsigned GetCount(float, const std::string&, double);
    unsigned GetCount(float, const std::string&, const std::string&);
    unsigned GetCount(double, int, int);
    unsigned GetCount(double, int, unsigned);
    unsigned GetCount(double, int, float);
    unsigned GetCount(double, int, double);
    unsigned GetCount(double, int, const std::string&);
    unsigned GetCount(double, unsigned, int);
    unsigned GetCount(double, unsigned, unsigned);
    unsigned GetCount(double, unsigned, float);
    unsigned GetCount(double, unsigned, double);
    unsigned GetCount(double, unsigned, const std::string&);
    unsigned GetCount(double, float, int);
    unsigned GetCount(double, float, unsigned);
    unsigned GetCount(double, float, float);
    unsigned GetCount(double, float, double);
    unsigned GetCount(double, float, const std::string&);
    unsigned GetCount(double, double, int);
    unsigned GetCount(double, double, unsigned);
    unsigned GetCount(double, double, float);
    unsigned GetCount(double, double, double);
    unsigned GetCount(double, double, const std::string&);
    unsigned GetCount(double, const std::string&, int);
    unsigned GetCount(double, const std::string&, unsigned);
    unsigned GetCount(double, const std::string&, float);
    unsigned GetCount(double, const std::string&, double);
    unsigned GetCount(double, const std::string&, const std::string&);
    unsigned GetCount(const std::string&, int, int);
    unsigned GetCount(const std::string&, int, unsigned);
    unsigned GetCount(const std::string&, int, float);
    unsigned GetCount(const std::string&, int, double);
    unsigned GetCount(const std::string&, int, const std::string&);
    unsigned GetCount(const std::string&, unsigned, int);
    unsigned GetCount(const std::string&, unsigned, unsigned);
    unsigned GetCount(const std::string&, unsigned, float);
    unsigned GetCount(const std::string&, unsigned, double);
    unsigned GetCount(const std::string&, unsigned, const std::string&);
    unsigned GetCount(const std::string&, float, int);
    unsigned GetCount(const std::string&, float, unsigned);
    unsigned GetCount(const std::string&, float, float);
    unsigned GetCount(const std::string&, float, double);
    unsigned GetCount(const std::string&, float, const std::string&);
    unsigned GetCount(const std::string&, double, int);
    unsigned GetCount(const std::string&, double, unsigned);
    unsigned GetCount(const std::string&, double, float);
    unsigned GetCount(const std::string&, double, double);
    unsigned GetCount(const std::string&, double, const std::string&);
    unsigned GetCount(const std::string&, const std::string&, int);
    unsigned GetCount(const std::string&, const std::string&, unsigned);
    unsigned GetCount(const std::string&, const std::string&, float);
    unsigned GetCount(const std::string&, const std::string&, double);
    unsigned GetCount(const std::string&, const std::string&, const std::string&);

#endif   /* CDB_3_ARGS */

#ifdef CDB_4_ARGS

    unsigned GetCount(int, int, int, int);
    unsigned GetCount(int, int, int, unsigned);
    unsigned GetCount(int, int, int, float);
    unsigned GetCount(int, int, int, double);
    unsigned GetCount(int, int, int, const std::string&);
    unsigned GetCount(int, int, unsigned, int);
    unsigned GetCount(int, int, unsigned, unsigned);
    unsigned GetCount(int, int, unsigned, float);
    unsigned GetCount(int, int, unsigned, double);
    unsigned GetCount(int, int, unsigned, const std::string&);
    unsigned GetCount(int, int, float, int);
    unsigned GetCount(int, int, float, unsigned);
    unsigned GetCount(int, int, float, float);
    unsigned GetCount(int, int, float, double);
    unsigned GetCount(int, int, float, const std::string&);
    unsigned GetCount(int, int, double, int);
    unsigned GetCount(int, int, double, unsigned);
    unsigned GetCount(int, int, double, float);
    unsigned GetCount(int, int, double, double);
    unsigned GetCount(int, int, double, const std::string&);
    unsigned GetCount(int, int, const std::string&, int);
    unsigned GetCount(int, int, const std::string&, unsigned);
    unsigned GetCount(int, int, const std::string&, float);
    unsigned GetCount(int, int, const std::string&, double);
    unsigned GetCount(int, int, const std::string&, const std::string&);
    unsigned GetCount(int, unsigned, int, int);
    unsigned GetCount(int, unsigned, int, unsigned);
    unsigned GetCount(int, unsigned, int, float);
    unsigned GetCount(int, unsigned, int, double);
    unsigned GetCount(int, unsigned, int, const std::string&);
    unsigned GetCount(int, unsigned, unsigned, int);
    unsigned GetCount(int, unsigned, unsigned, unsigned);
    unsigned GetCount(int, unsigned, unsigned, float);
    unsigned GetCount(int, unsigned, unsigned, double);
    unsigned GetCount(int, unsigned, unsigned, const std::string&);
    unsigned GetCount(int, unsigned, float, int);
    unsigned GetCount(int, unsigned, float, unsigned);
    unsigned GetCount(int, unsigned, float, float);
    unsigned GetCount(int, unsigned, float, double);
    unsigned GetCount(int, unsigned, float, const std::string&);
    unsigned GetCount(int, unsigned, double, int);
    unsigned GetCount(int, unsigned, double, unsigned);
    unsigned GetCount(int, unsigned, double, float);
    unsigned GetCount(int, unsigned, double, double);
    unsigned GetCount(int, unsigned, double, const std::string&);
    unsigned GetCount(int, unsigned, const std::string&, int);
    unsigned GetCount(int, unsigned, const std::string&, unsigned);
    unsigned GetCount(int, unsigned, const std::string&, float);
    unsigned GetCount(int, unsigned, const std::string&, double);
    unsigned GetCount(int, unsigned, const std::string&, const std::string&);
    unsigned GetCount(int, float, int, int);
    unsigned GetCount(int, float, int, unsigned);
    unsigned GetCount(int, float, int, float);
    unsigned GetCount(int, float, int, double);
    unsigned GetCount(int, float, int, const std::string&);
    unsigned GetCount(int, float, unsigned, int);
    unsigned GetCount(int, float, unsigned, unsigned);
    unsigned GetCount(int, float, unsigned, float);
    unsigned GetCount(int, float, unsigned, double);
    unsigned GetCount(int, float, unsigned, const std::string&);
    unsigned GetCount(int, float, float, int);
    unsigned GetCount(int, float, float, unsigned);
    unsigned GetCount(int, float, float, float);
    unsigned GetCount(int, float, float, double);
    unsigned GetCount(int, float, float, const std::string&);
    unsigned GetCount(int, float, double, int);
    unsigned GetCount(int, float, double, unsigned);
    unsigned GetCount(int, float, double, float);
    unsigned GetCount(int, float, double, double);
    unsigned GetCount(int, float, double, const std::string&);
    unsigned GetCount(int, float, const std::string&, int);
    unsigned GetCount(int, float, const std::string&, unsigned);
    unsigned GetCount(int, float, const std::string&, float);
    unsigned GetCount(int, float, const std::string&, double);
    unsigned GetCount(int, float, const std::string&, const std::string&);
    unsigned GetCount(int, double, int, int);
    unsigned GetCount(int, double, int, unsigned);
    unsigned GetCount(int, double, int, float);
    unsigned GetCount(int, double, int, double);
    unsigned GetCount(int, double, int, const std::string&);
    unsigned GetCount(int, double, unsigned, int);
    unsigned GetCount(int, double, unsigned, unsigned);
    unsigned GetCount(int, double, unsigned, float);
    unsigned GetCount(int, double, unsigned, double);
    unsigned GetCount(int, double, unsigned, const std::string&);
    unsigned GetCount(int, double, float, int);
    unsigned GetCount(int, double, float, unsigned);
    unsigned GetCount(int, double, float, float);
    unsigned GetCount(int, double, float, double);
    unsigned GetCount(int, double, float, const std::string&);
    unsigned GetCount(int, double, double, int);
    unsigned GetCount(int, double, double, unsigned);
    unsigned GetCount(int, double, double, float);
    unsigned GetCount(int, double, double, double);
    unsigned GetCount(int, double, double, const std::string&);
    unsigned GetCount(int, double, const std::string&, int);
    unsigned GetCount(int, double, const std::string&, unsigned);
    unsigned GetCount(int, double, const std::string&, float);
    unsigned GetCount(int, double, const std::string&, double);
    unsigned GetCount(int, double, const std::string&, const std::string&);
    unsigned GetCount(int, const std::string&, int, int);
    unsigned GetCount(int, const std::string&, int, unsigned);
    unsigned GetCount(int, const std::string&, int, float);
    unsigned GetCount(int, const std::string&, int, double);
    unsigned GetCount(int, const std::string&, int, const std::string&);
    unsigned GetCount(int, const std::string&, unsigned, int);
    unsigned GetCount(int, const std::string&, unsigned, unsigned);
    unsigned GetCount(int, const std::string&, unsigned, float);
    unsigned GetCount(int, const std::string&, unsigned, double);
    unsigned GetCount(int, const std::string&, unsigned, const std::string&);
    unsigned GetCount(int, const std::string&, float, int);
    unsigned GetCount(int, const std::string&, float, unsigned);
    unsigned GetCount(int, const std::string&, float, float);
    unsigned GetCount(int, const std::string&, float, double);
    unsigned GetCount(int, const std::string&, float, const std::string&);
    unsigned GetCount(int, const std::string&, double, int);
    unsigned GetCount(int, const std::string&, double, unsigned);
    unsigned GetCount(int, const std::string&, double, float);
    unsigned GetCount(int, const std::string&, double, double);
    unsigned GetCount(int, const std::string&, double, const std::string&);
    unsigned GetCount(int, const std::string&, const std::string&, int);
    unsigned GetCount(int, const std::string&, const std::string&, unsigned);
    unsigned GetCount(int, const std::string&, const std::string&, float);
    unsigned GetCount(int, const std::string&, const std::string&, double);
    unsigned GetCount(int, const std::string&, const std::string&, const std::string&);
    unsigned GetCount(unsigned, int, int, int);
    unsigned GetCount(unsigned, int, int, unsigned);
    unsigned GetCount(unsigned, int, int, float);
    unsigned GetCount(unsigned, int, int, double);
    unsigned GetCount(unsigned, int, int, const std::string&);
    unsigned GetCount(unsigned, int, unsigned, int);
    unsigned GetCount(unsigned, int, unsigned, unsigned);
    unsigned GetCount(unsigned, int, unsigned, float);
    unsigned GetCount(unsigned, int, unsigned, double);
    unsigned GetCount(unsigned, int, unsigned, const std::string&);
    unsigned GetCount(unsigned, int, float, int);
    unsigned GetCount(unsigned, int, float, unsigned);
    unsigned GetCount(unsigned, int, float, float);
    unsigned GetCount(unsigned, int, float, double);
    unsigned GetCount(unsigned, int, float, const std::string&);
    unsigned GetCount(unsigned, int, double, int);
    unsigned GetCount(unsigned, int, double, unsigned);
    unsigned GetCount(unsigned, int, double, float);
    unsigned GetCount(unsigned, int, double, double);
    unsigned GetCount(unsigned, int, double, const std::string&);
    unsigned GetCount(unsigned, int, const std::string&, int);
    unsigned GetCount(unsigned, int, const std::string&, unsigned);
    unsigned GetCount(unsigned, int, const std::string&, float);
    unsigned GetCount(unsigned, int, const std::string&, double);
    unsigned GetCount(unsigned, int, const std::string&, const std::string&);
    unsigned GetCount(unsigned, unsigned, int, int);
    unsigned GetCount(unsigned, unsigned, int, unsigned);
    unsigned GetCount(unsigned, unsigned, int, float);
    unsigned GetCount(unsigned, unsigned, int, double);
    unsigned GetCount(unsigned, unsigned, int, const std::string&);
    unsigned GetCount(unsigned, unsigned, unsigned, int);
    unsigned GetCount(unsigned, unsigned, unsigned, unsigned);
    unsigned GetCount(unsigned, unsigned, unsigned, float);
    unsigned GetCount(unsigned, unsigned, unsigned, double);
    unsigned GetCount(unsigned, unsigned, unsigned, const std::string&);
    unsigned GetCount(unsigned, unsigned, float, int);
    unsigned GetCount(unsigned, unsigned, float, unsigned);
    unsigned GetCount(unsigned, unsigned, float, float);
    unsigned GetCount(unsigned, unsigned, float, double);
    unsigned GetCount(unsigned, unsigned, float, const std::string&);
    unsigned GetCount(unsigned, unsigned, double, int);
    unsigned GetCount(unsigned, unsigned, double, unsigned);
    unsigned GetCount(unsigned, unsigned, double, float);
    unsigned GetCount(unsigned, unsigned, double, double);
    unsigned GetCount(unsigned, unsigned, double, const std::string&);
    unsigned GetCount(unsigned, unsigned, const std::string&, int);
    unsigned GetCount(unsigned, unsigned, const std::string&, unsigned);
    unsigned GetCount(unsigned, unsigned, const std::string&, float);
    unsigned GetCount(unsigned, unsigned, const std::string&, double);
    unsigned GetCount(unsigned, unsigned, const std::string&, const std::string&);
    unsigned GetCount(unsigned, float, int, int);
    unsigned GetCount(unsigned, float, int, unsigned);
    unsigned GetCount(unsigned, float, int, float);
    unsigned GetCount(unsigned, float, int, double);
    unsigned GetCount(unsigned, float, int, const std::string&);
    unsigned GetCount(unsigned, float, unsigned, int);
    unsigned GetCount(unsigned, float, unsigned, unsigned);
    unsigned GetCount(unsigned, float, unsigned, float);
    unsigned GetCount(unsigned, float, unsigned, double);
    unsigned GetCount(unsigned, float, unsigned, const std::string&);
    unsigned GetCount(unsigned, float, float, int);
    unsigned GetCount(unsigned, float, float, unsigned);
    unsigned GetCount(unsigned, float, float, float);
    unsigned GetCount(unsigned, float, float, double);
    unsigned GetCount(unsigned, float, float, const std::string&);
    unsigned GetCount(unsigned, float, double, int);
    unsigned GetCount(unsigned, float, double, unsigned);
    unsigned GetCount(unsigned, float, double, float);
    unsigned GetCount(unsigned, float, double, double);
    unsigned GetCount(unsigned, float, double, const std::string&);
    unsigned GetCount(unsigned, float, const std::string&, int);
    unsigned GetCount(unsigned, float, const std::string&, unsigned);
    unsigned GetCount(unsigned, float, const std::string&, float);
    unsigned GetCount(unsigned, float, const std::string&, double);
    unsigned GetCount(unsigned, float, const std::string&, const std::string&);
    unsigned GetCount(unsigned, double, int, int);
    unsigned GetCount(unsigned, double, int, unsigned);
    unsigned GetCount(unsigned, double, int, float);
    unsigned GetCount(unsigned, double, int, double);
    unsigned GetCount(unsigned, double, int, const std::string&);
    unsigned GetCount(unsigned, double, unsigned, int);
    unsigned GetCount(unsigned, double, unsigned, unsigned);
    unsigned GetCount(unsigned, double, unsigned, float);
    unsigned GetCount(unsigned, double, unsigned, double);
    unsigned GetCount(unsigned, double, unsigned, const std::string&);
    unsigned GetCount(unsigned, double, float, int);
    unsigned GetCount(unsigned, double, float, unsigned);
    unsigned GetCount(unsigned, double, float, float);
    unsigned GetCount(unsigned, double, float, double);
    unsigned GetCount(unsigned, double, float, const std::string&);
    unsigned GetCount(unsigned, double, double, int);
    unsigned GetCount(unsigned, double, double, unsigned);
    unsigned GetCount(unsigned, double, double, float);
    unsigned GetCount(unsigned, double, double, double);
    unsigned GetCount(unsigned, double, double, const std::string&);
    unsigned GetCount(unsigned, double, const std::string&, int);
    unsigned GetCount(unsigned, double, const std::string&, unsigned);
    unsigned GetCount(unsigned, double, const std::string&, float);
    unsigned GetCount(unsigned, double, const std::string&, double);
    unsigned GetCount(unsigned, double, const std::string&, const std::string&);
    unsigned GetCount(unsigned, const std::string&, int, int);
    unsigned GetCount(unsigned, const std::string&, int, unsigned);
    unsigned GetCount(unsigned, const std::string&, int, float);
    unsigned GetCount(unsigned, const std::string&, int, double);
    unsigned GetCount(unsigned, const std::string&, int, const std::string&);
    unsigned GetCount(unsigned, const std::string&, unsigned, int);
    unsigned GetCount(unsigned, const std::string&, unsigned, unsigned);
    unsigned GetCount(unsigned, const std::string&, unsigned, float);
    unsigned GetCount(unsigned, const std::string&, unsigned, double);
    unsigned GetCount(unsigned, const std::string&, unsigned, const std::string&);
    unsigned GetCount(unsigned, const std::string&, float, int);
    unsigned GetCount(unsigned, const std::string&, float, unsigned);
    unsigned GetCount(unsigned, const std::string&, float, float);
    unsigned GetCount(unsigned, const std::string&, float, double);
    unsigned GetCount(unsigned, const std::string&, float, const std::string&);
    unsigned GetCount(unsigned, const std::string&, double, int);
    unsigned GetCount(unsigned, const std::string&, double, unsigned);
    unsigned GetCount(unsigned, const std::string&, double, float);
    unsigned GetCount(unsigned, const std::string&, double, double);
    unsigned GetCount(unsigned, const std::string&, double, const std::string&);
    unsigned GetCount(unsigned, const std::string&, const std::string&, int);
    unsigned GetCount(unsigned, const std::string&, const std::string&, unsigned);
    unsigned GetCount(unsigned, const std::string&, const std::string&, float);
    unsigned GetCount(unsigned, const std::string&, const std::string&, double);
    unsigned GetCount(unsigned, const std::string&, const std::string&, const std::string&);
    unsigned GetCount(float, int, int, int);
    unsigned GetCount(float, int, int, unsigned);
    unsigned GetCount(float, int, int, float);
    unsigned GetCount(float, int, int, double);
    unsigned GetCount(float, int, int, const std::string&);
    unsigned GetCount(float, int, unsigned, int);
    unsigned GetCount(float, int, unsigned, unsigned);
    unsigned GetCount(float, int, unsigned, float);
    unsigned GetCount(float, int, unsigned, double);
    unsigned GetCount(float, int, unsigned, const std::string&);
    unsigned GetCount(float, int, float, int);
    unsigned GetCount(float, int, float, unsigned);
    unsigned GetCount(float, int, float, float);
    unsigned GetCount(float, int, float, double);
    unsigned GetCount(float, int, float, const std::string&);
    unsigned GetCount(float, int, double, int);
    unsigned GetCount(float, int, double, unsigned);
    unsigned GetCount(float, int, double, float);
    unsigned GetCount(float, int, double, double);
    unsigned GetCount(float, int, double, const std::string&);
    unsigned GetCount(float, int, const std::string&, int);
    unsigned GetCount(float, int, const std::string&, unsigned);
    unsigned GetCount(float, int, const std::string&, float);
    unsigned GetCount(float, int, const std::string&, double);
    unsigned GetCount(float, int, const std::string&, const std::string&);
    unsigned GetCount(float, unsigned, int, int);
    unsigned GetCount(float, unsigned, int, unsigned);
    unsigned GetCount(float, unsigned, int, float);
    unsigned GetCount(float, unsigned, int, double);
    unsigned GetCount(float, unsigned, int, const std::string&);
    unsigned GetCount(float, unsigned, unsigned, int);
    unsigned GetCount(float, unsigned, unsigned, unsigned);
    unsigned GetCount(float, unsigned, unsigned, float);
    unsigned GetCount(float, unsigned, unsigned, double);
    unsigned GetCount(float, unsigned, unsigned, const std::string&);
    unsigned GetCount(float, unsigned, float, int);
    unsigned GetCount(float, unsigned, float, unsigned);
    unsigned GetCount(float, unsigned, float, float);
    unsigned GetCount(float, unsigned, float, double);
    unsigned GetCount(float, unsigned, float, const std::string&);
    unsigned GetCount(float, unsigned, double, int);
    unsigned GetCount(float, unsigned, double, unsigned);
    unsigned GetCount(float, unsigned, double, float);
    unsigned GetCount(float, unsigned, double, double);
    unsigned GetCount(float, unsigned, double, const std::string&);
    unsigned GetCount(float, unsigned, const std::string&, int);
    unsigned GetCount(float, unsigned, const std::string&, unsigned);
    unsigned GetCount(float, unsigned, const std::string&, float);
    unsigned GetCount(float, unsigned, const std::string&, double);
    unsigned GetCount(float, unsigned, const std::string&, const std::string&);
    unsigned GetCount(float, float, int, int);
    unsigned GetCount(float, float, int, unsigned);
    unsigned GetCount(float, float, int, float);
    unsigned GetCount(float, float, int, double);
    unsigned GetCount(float, float, int, const std::string&);
    unsigned GetCount(float, float, unsigned, int);
    unsigned GetCount(float, float, unsigned, unsigned);
    unsigned GetCount(float, float, unsigned, float);
    unsigned GetCount(float, float, unsigned, double);
    unsigned GetCount(float, float, unsigned, const std::string&);
    unsigned GetCount(float, float, float, int);
    unsigned GetCount(float, float, float, unsigned);
    unsigned GetCount(float, float, float, float);
    unsigned GetCount(float, float, float, double);
    unsigned GetCount(float, float, float, const std::string&);
    unsigned GetCount(float, float, double, int);
    unsigned GetCount(float, float, double, unsigned);
    unsigned GetCount(float, float, double, float);
    unsigned GetCount(float, float, double, double);
    unsigned GetCount(float, float, double, const std::string&);
    unsigned GetCount(float, float, const std::string&, int);
    unsigned GetCount(float, float, const std::string&, unsigned);
    unsigned GetCount(float, float, const std::string&, float);
    unsigned GetCount(float, float, const std::string&, double);
    unsigned GetCount(float, float, const std::string&, const std::string&);
    unsigned GetCount(float, double, int, int);
    unsigned GetCount(float, double, int, unsigned);
    unsigned GetCount(float, double, int, float);
    unsigned GetCount(float, double, int, double);
    unsigned GetCount(float, double, int, const std::string&);
    unsigned GetCount(float, double, unsigned, int);
    unsigned GetCount(float, double, unsigned, unsigned);
    unsigned GetCount(float, double, unsigned, float);
    unsigned GetCount(float, double, unsigned, double);
    unsigned GetCount(float, double, unsigned, const std::string&);
    unsigned GetCount(float, double, float, int);
    unsigned GetCount(float, double, float, unsigned);
    unsigned GetCount(float, double, float, float);
    unsigned GetCount(float, double, float, double);
    unsigned GetCount(float, double, float, const std::string&);
    unsigned GetCount(float, double, double, int);
    unsigned GetCount(float, double, double, unsigned);
    unsigned GetCount(float, double, double, float);
    unsigned GetCount(float, double, double, double);
    unsigned GetCount(float, double, double, const std::string&);
    unsigned GetCount(float, double, const std::string&, int);
    unsigned GetCount(float, double, const std::string&, unsigned);
    unsigned GetCount(float, double, const std::string&, float);
    unsigned GetCount(float, double, const std::string&, double);
    unsigned GetCount(float, double, const std::string&, const std::string&);
    unsigned GetCount(float, const std::string&, int, int);
    unsigned GetCount(float, const std::string&, int, unsigned);
    unsigned GetCount(float, const std::string&, int, float);
    unsigned GetCount(float, const std::string&, int, double);
    unsigned GetCount(float, const std::string&, int, const std::string&);
    unsigned GetCount(float, const std::string&, unsigned, int);
    unsigned GetCount(float, const std::string&, unsigned, unsigned);
    unsigned GetCount(float, const std::string&, unsigned, float);
    unsigned GetCount(float, const std::string&, unsigned, double);
    unsigned GetCount(float, const std::string&, unsigned, const std::string&);
    unsigned GetCount(float, const std::string&, float, int);
    unsigned GetCount(float, const std::string&, float, unsigned);
    unsigned GetCount(float, const std::string&, float, float);
    unsigned GetCount(float, const std::string&, float, double);
    unsigned GetCount(float, const std::string&, float, const std::string&);
    unsigned GetCount(float, const std::string&, double, int);
    unsigned GetCount(float, const std::string&, double, unsigned);
    unsigned GetCount(float, const std::string&, double, float);
    unsigned GetCount(float, const std::string&, double, double);
    unsigned GetCount(float, const std::string&, double, const std::string&);
    unsigned GetCount(float, const std::string&, const std::string&, int);
    unsigned GetCount(float, const std::string&, const std::string&, unsigned);
    unsigned GetCount(float, const std::string&, const std::string&, float);
    unsigned GetCount(float, const std::string&, const std::string&, double);
    unsigned GetCount(float, const std::string&, const std::string&, const std::string&);
    unsigned GetCount(double, int, int, int);
    unsigned GetCount(double, int, int, unsigned);
    unsigned GetCount(double, int, int, float);
    unsigned GetCount(double, int, int, double);
    unsigned GetCount(double, int, int, const std::string&);
    unsigned GetCount(double, int, unsigned, int);
    unsigned GetCount(double, int, unsigned, unsigned);
    unsigned GetCount(double, int, unsigned, float);
    unsigned GetCount(double, int, unsigned, double);
    unsigned GetCount(double, int, unsigned, const std::string&);
    unsigned GetCount(double, int, float, int);
    unsigned GetCount(double, int, float, unsigned);
    unsigned GetCount(double, int, float, float);
    unsigned GetCount(double, int, float, double);
    unsigned GetCount(double, int, float, const std::string&);
    unsigned GetCount(double, int, double, int);
    unsigned GetCount(double, int, double, unsigned);
    unsigned GetCount(double, int, double, float);
    unsigned GetCount(double, int, double, double);
    unsigned GetCount(double, int, double, const std::string&);
    unsigned GetCount(double, int, const std::string&, int);
    unsigned GetCount(double, int, const std::string&, unsigned);
    unsigned GetCount(double, int, const std::string&, float);
    unsigned GetCount(double, int, const std::string&, double);
    unsigned GetCount(double, int, const std::string&, const std::string&);
    unsigned GetCount(double, unsigned, int, int);
    unsigned GetCount(double, unsigned, int, unsigned);
    unsigned GetCount(double, unsigned, int, float);
    unsigned GetCount(double, unsigned, int, double);
    unsigned GetCount(double, unsigned, int, const std::string&);
    unsigned GetCount(double, unsigned, unsigned, int);
    unsigned GetCount(double, unsigned, unsigned, unsigned);
    unsigned GetCount(double, unsigned, unsigned, float);
    unsigned GetCount(double, unsigned, unsigned, double);
    unsigned GetCount(double, unsigned, unsigned, const std::string&);
    unsigned GetCount(double, unsigned, float, int);
    unsigned GetCount(double, unsigned, float, unsigned);
    unsigned GetCount(double, unsigned, float, float);
    unsigned GetCount(double, unsigned, float, double);
    unsigned GetCount(double, unsigned, float, const std::string&);
    unsigned GetCount(double, unsigned, double, int);
    unsigned GetCount(double, unsigned, double, unsigned);
    unsigned GetCount(double, unsigned, double, float);
    unsigned GetCount(double, unsigned, double, double);
    unsigned GetCount(double, unsigned, double, const std::string&);
    unsigned GetCount(double, unsigned, const std::string&, int);
    unsigned GetCount(double, unsigned, const std::string&, unsigned);
    unsigned GetCount(double, unsigned, const std::string&, float);
    unsigned GetCount(double, unsigned, const std::string&, double);
    unsigned GetCount(double, unsigned, const std::string&, const std::string&);
    unsigned GetCount(double, float, int, int);
    unsigned GetCount(double, float, int, unsigned);
    unsigned GetCount(double, float, int, float);
    unsigned GetCount(double, float, int, double);
    unsigned GetCount(double, float, int, const std::string&);
    unsigned GetCount(double, float, unsigned, int);
    unsigned GetCount(double, float, unsigned, unsigned);
    unsigned GetCount(double, float, unsigned, float);
    unsigned GetCount(double, float, unsigned, double);
    unsigned GetCount(double, float, unsigned, const std::string&);
    unsigned GetCount(double, float, float, int);
    unsigned GetCount(double, float, float, unsigned);
    unsigned GetCount(double, float, float, float);
    unsigned GetCount(double, float, float, double);
    unsigned GetCount(double, float, float, const std::string&);
    unsigned GetCount(double, float, double, int);
    unsigned GetCount(double, float, double, unsigned);
    unsigned GetCount(double, float, double, float);
    unsigned GetCount(double, float, double, double);
    unsigned GetCount(double, float, double, const std::string&);
    unsigned GetCount(double, float, const std::string&, int);
    unsigned GetCount(double, float, const std::string&, unsigned);
    unsigned GetCount(double, float, const std::string&, float);
    unsigned GetCount(double, float, const std::string&, double);
    unsigned GetCount(double, float, const std::string&, const std::string&);
    unsigned GetCount(double, double, int, int);
    unsigned GetCount(double, double, int, unsigned);
    unsigned GetCount(double, double, int, float);
    unsigned GetCount(double, double, int, double);
    unsigned GetCount(double, double, int, const std::string&);
    unsigned GetCount(double, double, unsigned, int);
    unsigned GetCount(double, double, unsigned, unsigned);
    unsigned GetCount(double, double, unsigned, float);
    unsigned GetCount(double, double, unsigned, double);
    unsigned GetCount(double, double, unsigned, const std::string&);
    unsigned GetCount(double, double, float, int);
    unsigned GetCount(double, double, float, unsigned);
    unsigned GetCount(double, double, float, float);
    unsigned GetCount(double, double, float, double);
    unsigned GetCount(double, double, float, const std::string&);
    unsigned GetCount(double, double, double, int);
    unsigned GetCount(double, double, double, unsigned);
    unsigned GetCount(double, double, double, float);
    unsigned GetCount(double, double, double, double);
    unsigned GetCount(double, double, double, const std::string&);
    unsigned GetCount(double, double, const std::string&, int);
    unsigned GetCount(double, double, const std::string&, unsigned);
    unsigned GetCount(double, double, const std::string&, float);
    unsigned GetCount(double, double, const std::string&, double);
    unsigned GetCount(double, double, const std::string&, const std::string&);
    unsigned GetCount(double, const std::string&, int, int);
    unsigned GetCount(double, const std::string&, int, unsigned);
    unsigned GetCount(double, const std::string&, int, float);
    unsigned GetCount(double, const std::string&, int, double);
    unsigned GetCount(double, const std::string&, int, const std::string&);
    unsigned GetCount(double, const std::string&, unsigned, int);
    unsigned GetCount(double, const std::string&, unsigned, unsigned);
    unsigned GetCount(double, const std::string&, unsigned, float);
    unsigned GetCount(double, const std::string&, unsigned, double);
    unsigned GetCount(double, const std::string&, unsigned, const std::string&);
    unsigned GetCount(double, const std::string&, float, int);
    unsigned GetCount(double, const std::string&, float, unsigned);
    unsigned GetCount(double, const std::string&, float, float);
    unsigned GetCount(double, const std::string&, float, double);
    unsigned GetCount(double, const std::string&, float, const std::string&);
    unsigned GetCount(double, const std::string&, double, int);
    unsigned GetCount(double, const std::string&, double, unsigned);
    unsigned GetCount(double, const std::string&, double, float);
    unsigned GetCount(double, const std::string&, double, double);
    unsigned GetCount(double, const std::string&, double, const std::string&);
    unsigned GetCount(double, const std::string&, const std::string&, int);
    unsigned GetCount(double, const std::string&, const std::string&, unsigned);
    unsigned GetCount(double, const std::string&, const std::string&, float);
    unsigned GetCount(double, const std::string&, const std::string&, double);
    unsigned GetCount(double, const std::string&, const std::string&, const std::string&);
    unsigned GetCount(const std::string&, int, int, int);
    unsigned GetCount(const std::string&, int, int, unsigned);
    unsigned GetCount(const std::string&, int, int, float);
    unsigned GetCount(const std::string&, int, int, double);
    unsigned GetCount(const std::string&, int, int, const std::string&);
    unsigned GetCount(const std::string&, int, unsigned, int);
    unsigned GetCount(const std::string&, int, unsigned, unsigned);
    unsigned GetCount(const std::string&, int, unsigned, float);
    unsigned GetCount(const std::string&, int, unsigned, double);
    unsigned GetCount(const std::string&, int, unsigned, const std::string&);
    unsigned GetCount(const std::string&, int, float, int);
    unsigned GetCount(const std::string&, int, float, unsigned);
    unsigned GetCount(const std::string&, int, float, float);
    unsigned GetCount(const std::string&, int, float, double);
    unsigned GetCount(const std::string&, int, float, const std::string&);
    unsigned GetCount(const std::string&, int, double, int);
    unsigned GetCount(const std::string&, int, double, unsigned);
    unsigned GetCount(const std::string&, int, double, float);
    unsigned GetCount(const std::string&, int, double, double);
    unsigned GetCount(const std::string&, int, double, const std::string&);
    unsigned GetCount(const std::string&, int, const std::string&, int);
    unsigned GetCount(const std::string&, int, const std::string&, unsigned);
    unsigned GetCount(const std::string&, int, const std::string&, float);
    unsigned GetCount(const std::string&, int, const std::string&, double);
    unsigned GetCount(const std::string&, int, const std::string&, const std::string&);
    unsigned GetCount(const std::string&, unsigned, int, int);
    unsigned GetCount(const std::string&, unsigned, int, unsigned);
    unsigned GetCount(const std::string&, unsigned, int, float);
    unsigned GetCount(const std::string&, unsigned, int, double);
    unsigned GetCount(const std::string&, unsigned, int, const std::string&);
    unsigned GetCount(const std::string&, unsigned, unsigned, int);
    unsigned GetCount(const std::string&, unsigned, unsigned, unsigned);
    unsigned GetCount(const std::string&, unsigned, unsigned, float);
    unsigned GetCount(const std::string&, unsigned, unsigned, double);
    unsigned GetCount(const std::string&, unsigned, unsigned, const std::string&);
    unsigned GetCount(const std::string&, unsigned, float, int);
    unsigned GetCount(const std::string&, unsigned, float, unsigned);
    unsigned GetCount(const std::string&, unsigned, float, float);
    unsigned GetCount(const std::string&, unsigned, float, double);
    unsigned GetCount(const std::string&, unsigned, float, const std::string&);
    unsigned GetCount(const std::string&, unsigned, double, int);
    unsigned GetCount(const std::string&, unsigned, double, unsigned);
    unsigned GetCount(const std::string&, unsigned, double, float);
    unsigned GetCount(const std::string&, unsigned, double, double);
    unsigned GetCount(const std::string&, unsigned, double, const std::string&);
    unsigned GetCount(const std::string&, unsigned, const std::string&, int);
    unsigned GetCount(const std::string&, unsigned, const std::string&, unsigned);
    unsigned GetCount(const std::string&, unsigned, const std::string&, float);
    unsigned GetCount(const std::string&, unsigned, const std::string&, double);
    unsigned GetCount(const std::string&, unsigned, const std::string&, const std::string&);
    unsigned GetCount(const std::string&, float, int, int);
    unsigned GetCount(const std::string&, float, int, unsigned);
    unsigned GetCount(const std::string&, float, int, float);
    unsigned GetCount(const std::string&, float, int, double);
    unsigned GetCount(const std::string&, float, int, const std::string&);
    unsigned GetCount(const std::string&, float, unsigned, int);
    unsigned GetCount(const std::string&, float, unsigned, unsigned);
    unsigned GetCount(const std::string&, float, unsigned, float);
    unsigned GetCount(const std::string&, float, unsigned, double);
    unsigned GetCount(const std::string&, float, unsigned, const std::string&);
    unsigned GetCount(const std::string&, float, float, int);
    unsigned GetCount(const std::string&, float, float, unsigned);
    unsigned GetCount(const std::string&, float, float, float);
    unsigned GetCount(const std::string&, float, float, double);
    unsigned GetCount(const std::string&, float, float, const std::string&);
    unsigned GetCount(const std::string&, float, double, int);
    unsigned GetCount(const std::string&, float, double, unsigned);
    unsigned GetCount(const std::string&, float, double, float);
    unsigned GetCount(const std::string&, float, double, double);
    unsigned GetCount(const std::string&, float, double, const std::string&);
    unsigned GetCount(const std::string&, float, const std::string&, int);
    unsigned GetCount(const std::string&, float, const std::string&, unsigned);
    unsigned GetCount(const std::string&, float, const std::string&, float);
    unsigned GetCount(const std::string&, float, const std::string&, double);
    unsigned GetCount(const std::string&, float, const std::string&, const std::string&);
    unsigned GetCount(const std::string&, double, int, int);
    unsigned GetCount(const std::string&, double, int, unsigned);
    unsigned GetCount(const std::string&, double, int, float);
    unsigned GetCount(const std::string&, double, int, double);
    unsigned GetCount(const std::string&, double, int, const std::string&);
    unsigned GetCount(const std::string&, double, unsigned, int);
    unsigned GetCount(const std::string&, double, unsigned, unsigned);
    unsigned GetCount(const std::string&, double, unsigned, float);
    unsigned GetCount(const std::string&, double, unsigned, double);
    unsigned GetCount(const std::string&, double, unsigned, const std::string&);
    unsigned GetCount(const std::string&, double, float, int);
    unsigned GetCount(const std::string&, double, float, unsigned);
    unsigned GetCount(const std::string&, double, float, float);
    unsigned GetCount(const std::string&, double, float, double);
    unsigned GetCount(const std::string&, double, float, const std::string&);
    unsigned GetCount(const std::string&, double, double, int);
    unsigned GetCount(const std::string&, double, double, unsigned);
    unsigned GetCount(const std::string&, double, double, float);
    unsigned GetCount(const std::string&, double, double, double);
    unsigned GetCount(const std::string&, double, double, const std::string&);
    unsigned GetCount(const std::string&, double, const std::string&, int);
    unsigned GetCount(const std::string&, double, const std::string&, unsigned);
    unsigned GetCount(const std::string&, double, const std::string&, float);
    unsigned GetCount(const std::string&, double, const std::string&, double);
    unsigned GetCount(const std::string&, double, const std::string&, const std::string&);
    unsigned GetCount(const std::string&, const std::string&, int, int);
    unsigned GetCount(const std::string&, const std::string&, int, unsigned);
    unsigned GetCount(const std::string&, const std::string&, int, float);
    unsigned GetCount(const std::string&, const std::string&, int, double);
    unsigned GetCount(const std::string&, const std::string&, int, const std::string&);
    unsigned GetCount(const std::string&, const std::string&, unsigned, int);
    unsigned GetCount(const std::string&, const std::string&, unsigned, unsigned);
    unsigned GetCount(const std::string&, const std::string&, unsigned, float);
    unsigned GetCount(const std::string&, const std::string&, unsigned, double);
    unsigned GetCount(const std::string&, const std::string&, unsigned, const std::string&);
    unsigned GetCount(const std::string&, const std::string&, float, int);
    unsigned GetCount(const std::string&, const std::string&, float, unsigned);
    unsigned GetCount(const std::string&, const std::string&, float, float);
    unsigned GetCount(const std::string&, const std::string&, float, double);
    unsigned GetCount(const std::string&, const std::string&, float, const std::string&);
    unsigned GetCount(const std::string&, const std::string&, double, int);
    unsigned GetCount(const std::string&, const std::string&, double, unsigned);
    unsigned GetCount(const std::string&, const std::string&, double, float);
    unsigned GetCount(const std::string&, const std::string&, double, double);
    unsigned GetCount(const std::string&, const std::string&, double, const std::string&);
    unsigned GetCount(const std::string&, const std::string&, const std::string&, int);
    unsigned GetCount(const std::string&, const std::string&, const std::string&, unsigned);
    unsigned GetCount(const std::string&, const std::string&, const std::string&, float);
    unsigned GetCount(const std::string&, const std::string&, const std::string&, double);
    unsigned GetCount(const std::string&, const std::string&, const std::string&, const std::string&);

#endif   /* CDB_4_ARGS */


protected:

static MYSQL    *dbConn_;

    int         instance_;

    std::string table_;
    std::string columnList_;

    unsigned    skip_;
    unsigned    first_;

    MYSQL_STMT  *s_;
    MYSQL_STMT  *sc_;
    MYSQL_STMT  *sGet_;
    bool        firstGet_;
    MYSQL_STMT  *sUpdate_;
    bool        firstUpdate_;
    MYSQL_STMT  *sInsert_;
    bool        firstInsert_;
    MYSQL_STMT  *sDelete_;
    bool        firstDelete_;
    MYSQL_STMT  *sSet_;
    bool        firstSet_;

    MYSQL_BIND  bnd_[16];
    MYSQL_BIND  bndci_[16];
    MYSQL_BIND  bndco_[1];
    MYSQL_BIND  bndso_[1];

    void setInstance(bool *slots);
    void openCursor(MYSQL_STMT *s);
    unsigned getCount(MYSQL_STMT *s);
    void set_time(MYSQL_TIME *ts, const char *str);
    void set_date(MYSQL_TIME *ts, const char *str);
    void set_datetime(MYSQL_TIME *ts, const char *str);
    virtual void bindOutput(MYSQL_STMT *s) = 0;
    virtual void bindSetOutput() = 0;


private:

static std::string dbName_;
static bool open_;
};


#endif /* NPP_MYSQL_H */
